###################################################
### chunk number 1: set-opts
###################################################
  glop <- options(keep.source=TRUE,width=60,continue=" ",prompt=" ")
  library(pomp)
  pdf.options(useDingbats=FALSE)
  set.seed(5384959)


###################################################
### chunk number 2: ou2-proc-sim-def
###################################################
require(pomp)

ou2.proc.sim <- function (x, t, params, delta.t, ...) {
  xi <- rnorm(n=2,mean=0,sd=1) # noise terms
  xnew <- c(
            params["alpha.1"]*x["x1"]+params["alpha.3"]*x["x2"]+
            params["sigma.1"]*xi[1],
            params["alpha.2"]*x["x1"]+params["alpha.4"]*x["x2"]+
            params["sigma.2"]*xi[1]+params["sigma.3"]*xi[2]
            )
  names(xnew) <- c("x1","x2")
  xnew
}


###################################################
### chunk number 3: ou2-meas-sim-def
###################################################
ou2.meas.sim <- function (x, t, params, ...) {
  y <- rnorm(n=2,mean=x[c("x1","x2")],sd=params["tau"])
  names(y) <- c("y1","y2")
  y
}


###################################################
### chunk number 4: first-pomp-construction
###################################################
ou2 <- pomp(
            data=data.frame(
              time=1:100,
              y1=NA,
              y2=NA
              ),
            times="time",
            rprocess=discrete.time.sim(
              step.fun=ou2.proc.sim
              ),
            rmeasure=ou2.meas.sim,
            t0=0
            )


###################################################
### chunk number 5: set-params
###################################################
theta <- c(
           alpha.1=0.8, alpha.2=-0.5, alpha.3=0.3, alpha.4=0.9,
           sigma.1=3, sigma.2=-0.5, sigma.3=2,
           tau=1, 
           x1.0=-3, x2.0=4
           )


###################################################
### chunk number 6: ou2-first-simulation eval=FALSE
###################################################
## ou2 <- simulate(ou2,params=theta)


###################################################
### chunk number 7:  eval=FALSE
###################################################
## plot(ou2,variables=c("y1","y2"))


###################################################
### chunk number 8: ou2-plot
###################################################
data(ou2)
plot(ou2,variables=c("y1","y2"))


###################################################
### chunk number 9: ou2-meas-dens-def
###################################################
ou2.meas.dens <- function (y, x, t, params, log, ...) {
  f <- sum(
           dnorm(
                 x=y[c("y1","y2")],
                 mean=x[c("x1","x2")],
                 sd=params["tau"],
                 log=TRUE
                 ),
           na.rm=TRUE
           )
  if (log) f else exp(f)
}


###################################################
### chunk number 10: second-pomp-construction
###################################################
dat <- as(ou2,"data.frame")
theta <- coef(ou2)
ou2 <- pomp(
            data=dat[c("time","y1","y2")],
            times="time",
            rprocess=discrete.time.sim(ou2.proc.sim),
            rmeasure=ou2.meas.sim,
            dmeasure=ou2.meas.dens,
            t0=0
            )
coef(ou2) <- theta


###################################################
### chunk number 11: 
###################################################
rm(ou2)
data(ou2)


###################################################
### chunk number 12: ou2-pfilter-truth eval=FALSE
###################################################
## pf <- pfilter(ou2,params=theta,Np=1000)
## loglik.truth <- pf$loglik
## loglik.truth


###################################################
### chunk number 13: ou2-pfilter-truth-eval
###################################################
set.seed(457645443L)
pf <- pfilter(ou2,params=theta,Np=1000)
loglik.truth <- pf$loglik
loglik.truth


###################################################
### chunk number 14: ou2-pfilter-truth-alt1 eval=FALSE
###################################################
## pf <- pfilter(ou2,params=coef(ou2),Np=1000)


###################################################
### chunk number 15: ou2-pfilter-truth-alt2 eval=FALSE
###################################################
## pf <- pfilter(ou2,Np=1000)


###################################################
### chunk number 16: ou2-pfilter-guess eval=FALSE
###################################################
## theta.true <- coef(ou2)
## theta.guess <- theta.true
## theta.guess[c("alpha.2","alpha.3","tau")] <- 1.5*theta.true[c("alpha.2","alpha.3","tau")]
## pf <- pfilter(ou2,params=theta.guess,Np=1000)
## loglik.guess <- pf$loglik


###################################################
### chunk number 17: ou2-pfilter-guess-eval
###################################################
set.seed(457645443L)
theta.true <- coef(ou2)
theta.guess <- theta.true
theta.guess[c("alpha.2","alpha.3","tau")] <- 1.5*theta.true[c("alpha.2","alpha.3","tau")]
pf <- pfilter(ou2,params=theta.guess,Np=1000)
loglik.guess <- pf$loglik


###################################################
### chunk number 18: kalman-filter-def
###################################################
require(mvtnorm)
kalman.filter <- function (y, x0, a, b, sigma, tau) {
  n <- nrow(y)
  ntimes <- ncol(y)
  sigma.sq <- sigma%*%t(sigma)
  tau.sq <- tau%*%t(tau)
  inv.tau.sq <- solve(tau.sq)
  cond.loglik <- numeric(ntimes)
  filter.mean <- matrix(0,n,ntimes)
  pred.mean <- matrix(0,n,ntimes)
  pred.var <- array(0,dim=c(n,n,ntimes))
  m <- x0
  v <- diag(0,n)
  for (k in seq_len(ntimes)) {
    pred.mean[,k] <- M <- a%*%m
    pred.var[,,k] <- V <- a%*%v%*%t(a)+sigma.sq
    q <- b%*%V%*%t(b)+tau.sq
    r <- y[,k]-b%*%M
    cond.loglik[k] <- dmvnorm(x=y[,k],mean=as.numeric(b%*%M),sigma=q,log=TRUE)
    q <- t(b)%*%inv.tau.sq%*%b+solve(V)
    v <- solve(q)
    filter.mean[,k] <- m <- v%*%(t(b)%*%inv.tau.sq%*%y[,k]+solve(V,M))
  }
  list(
       pred.mean=pred.mean,
       pred.var=pred.var,
       filter.mean=filter.mean,
       cond.loglik=cond.loglik,
       loglik=sum(cond.loglik)
       )
}


###################################################
### chunk number 19: kalman-filter-run
###################################################
y <- data.array(ou2)
a <- matrix(theta.guess[c('alpha.1','alpha.2','alpha.3','alpha.4')],nrow=2,ncol=2)
b <- diag(1,2)   ## b is the identity matrix
sigma <- matrix(
                c(
                  theta.guess['sigma.1'],0,
                  theta.guess['sigma.2'],theta.guess['sigma.3']
                  ),
                nrow=2,ncol=2,byrow=T
                )
tau <- diag(theta.guess['tau'],2,2)
x0 <- init.state(ou2)
kf <- kalman.filter(y,x0,a,b,sigma,tau)


###################################################
### chunk number 20:  eval=FALSE
###################################################
## as(ou2,'data.frame')


###################################################
### chunk number 21:  eval=FALSE
###################################################
## data.array(ou2)
## time(ou2)  


###################################################
### chunk number 22:  eval=FALSE
###################################################
## time(ou2) <- 1:10


###################################################
### chunk number 23:  eval=FALSE
###################################################
## ou2 <- window(ou2,start=1,end=10)


###################################################
### chunk number 24:  eval=FALSE
###################################################
## timezero(ou2)
## timezero(ou2) <- -10


###################################################
### chunk number 25:  eval=FALSE
###################################################
## time(ou2,t0=TRUE)  
## time(ou2,t0=T) <- seq(from=0,to=10,by=1)


###################################################
### chunk number 26:  eval=FALSE
###################################################
## coef(ou2)
## coef(ou2,c("sigma.1","sigma.2")) <- c(1,0)


###################################################
### chunk number 27: ou2-multi-mif-calc eval=FALSE
###################################################
## mf <- replicate(
##                 n=3,
##                 mif(
##                     ou2,
##                     Nmif=120,
##                     start=theta.guess,
##                     pars=c('alpha.2','alpha.3','tau'),
##                     ivps=c('x1.0','x2.0'),
##                     rw.sd=c(
##                       x1.0=5,x2.0=5,
##                       alpha.2=0.02,alpha.3=0.02,tau=0.05
##                       ),
##                     Np=1000,
##                     var.factor=4,
##                     ic.lag=10,
##                     cooling.factor=0.97,
##                     max.fail=10
##                     )
##                 )
## fitted.pars <- c("alpha.2","alpha.3","tau","x1.0","x2.0")
## pf <- lapply(mf,pfilter)
## loglik.mle <- log(mean(exp(480+sapply(pf,function(x)x$loglik))))-480
## loglik.mle.sd <- sd(sapply(pf,function(x)x$loglik))
## theta.mle <- apply(sapply(mf,coef),1,mean)


###################################################
### chunk number 28: ou2-first-mif-calc eval=FALSE
###################################################
## mf <- mif(
##           ou2,
##           Nmif=120,
##           start=theta.guess,
##           pars=c('alpha.2','alpha.3','tau'),
##           ivps=c('x1.0','x2.0'),
##           rw.sd=c(
##             x1.0=5,x2.0=5,
##             alpha.2=0.02,alpha.3=0.02,tau=0.05
##             ),
##           Np=1000,
##           var.factor=4,
##           ic.lag=10,
##           cooling.factor=0.97,
##           max.fail=10
##           )
## fitted.pars <- c("alpha.2","alpha.3","tau","x1.0","x2.0")
## pf <- pfilter(mf)
## loglik.mle <- pf$loglik
## theta.mle <- coef(mf)


###################################################
### chunk number 29: ou2-first-mif-eval
###################################################
set.seed(334388458L)
binary.file <- "ou2-first-mif.rda"
if (file.exists(binary.file)) {
  load(binary.file)
} else {
  tic <- Sys.time()
mf <- replicate(
                n=3,
                mif(
                    ou2,
                    Nmif=120,
                    start=theta.guess,
                    pars=c('alpha.2','alpha.3','tau'),
                    ivps=c('x1.0','x2.0'),
                    rw.sd=c(
                      x1.0=5,x2.0=5,
                      alpha.2=0.02,alpha.3=0.02,tau=0.05
                      ),
                    Np=1000,
                    var.factor=4,
                    ic.lag=10,
                    cooling.factor=0.97,
                    max.fail=10
                    )
                )
fitted.pars <- c("alpha.2","alpha.3","tau","x1.0","x2.0")
pf <- lapply(mf,pfilter)
loglik.mle <- log(mean(exp(480+sapply(pf,function(x)x$loglik))))-480
loglik.mle.sd <- sd(sapply(pf,function(x)x$loglik))
theta.mle <- apply(sapply(mf,coef),1,mean)
  toc <- Sys.time()
  etime <- toc-tic
  save(mf,fitted.pars,theta.guess,theta.true,theta.mle,loglik.mle,etime,file=binary.file)
}
cbind(
      guess=c(signif(theta.guess[fitted.pars],3),loglik=round(loglik.guess,1)),
      mle=c(signif(theta.mle[fitted.pars],3),loglik=round(loglik.mle,1)),
      truth=c(signif(theta.true[fitted.pars],3),loglik=round(loglik.truth,1))
      ) -> results.table


###################################################
### chunk number 30: first-mif-results-table
###################################################
print(results.table)


###################################################
### chunk number 31: first-mif-plot eval=FALSE
###################################################
## op <- par(mfrow=c(4,1),mar=c(3,3,0,0),mgp=c(2,1,0),bty='l')
## x <- conv.rec(mf,c("loglik","alpha.2","alpha.3","tau"))
## plot(x[,"loglik"],type='l',xlab="",ylab=expression(paste(log,L)),xaxt='n')
## plot(x[,"alpha.2"],type='l',xlab="",ylab=expression(alpha[2]),xaxt='n')
## plot(x[,"alpha.3"],type='l',xlab="",ylab=expression(alpha[3]),xaxt='n')
## plot(x[,"tau"],type='l',xlab="MIF iteration",ylab=expression(tau))
## par(op)


###################################################
### chunk number 32: multi-mif-plot eval=FALSE
###################################################
## op <- par(mfrow=c(4,1),mar=c(3,3,0,0),mgp=c(2,1,0),bty='l')
## loglik <- sapply(mf,function(x)conv.rec(x,"loglik"))
## alpha.2 <- sapply(mf,function(x)conv.rec(x,"alpha.2"))
## alpha.3 <- sapply(mf,function(x)conv.rec(x,"alpha.3"))
## tau <- sapply(mf,function(x)conv.rec(x,"tau"))
## matplot(loglik,type='l',lty=1,xlab="",ylab=expression(paste(log,L)),xaxt='n')
## matplot(alpha.2,type='l',lty=1,xlab="",ylab=expression(alpha[2]),xaxt='n')
## matplot(alpha.3,type='l',lty=1,xlab="",ylab=expression(alpha[3]),xaxt='n')
## matplot(tau,type='l',lty=1,xlab="MIF iteration",ylab=expression(tau))
## par(op)


###################################################
### chunk number 33: mif-plot
###################################################
op <- par(mfrow=c(4,1),mar=c(3,3,0,0),mgp=c(2,1,0),bty='l')
loglik <- sapply(mf,function(x)conv.rec(x,"loglik"))
alpha.2 <- sapply(mf,function(x)conv.rec(x,"alpha.2"))
alpha.3 <- sapply(mf,function(x)conv.rec(x,"alpha.3"))
tau <- sapply(mf,function(x)conv.rec(x,"tau"))
matplot(loglik,type='l',lty=1,xlab="",ylab=expression(paste(log,L)),xaxt='n')
matplot(alpha.2,type='l',lty=1,xlab="",ylab=expression(alpha[2]),xaxt='n')
matplot(alpha.3,type='l',lty=1,xlab="",ylab=expression(alpha[3]),xaxt='n')
matplot(tau,type='l',lty=1,xlab="MIF iteration",ylab=expression(tau))
par(op)


###################################################
### chunk number 34: first-nlf eval=FALSE
###################################################
## estnames <- c('alpha.2','alpha.3','tau')
## out <- nlf(
##            ou2,
##            start=theta.guess,
##            nasymp=2000,
##            est=estnames,
##            lags=c(4,6),
##            seed=5669345L,
##            skip.se=TRUE,
##            method="Nelder-Mead",
##            trace=0,
##            maxit=100,
##            reltol=1e-8,
##            transform=function(x)x,
##            eval.only=FALSE
##            )


###################################################
### chunk number 35: first-nlf-results eval=FALSE
###################################################
## print(
##       cbind(
##             guess=theta.guess[estnames],
##             fit=out$params[estnames],
##             truth=theta.true[estnames]
##             ),
##       digits=3
##       )


###################################################
### chunk number 36: ou2-skeleton-def
###################################################
ou2.skel <- function (x, t, params, ...) {
  xnew <- c(
            params["alpha.1"]*x["x1"]+params["alpha.3"]*x["x2"],
            params["alpha.2"]*x["x1"]+params["alpha.4"]*x["x2"]
            )
  names(xnew) <- c("x1","x2")
  xnew
}


###################################################
### chunk number 37: new-ou2-pomp-construction
###################################################
dat <- subset(as(ou2,"data.frame"),time<=60)
theta <- coef(ou2)
new.ou2 <- pomp(
                data=dat[c("time","y1","y2")],
                times="time",
                rprocess=discrete.time.sim(ou2.proc.sim),
                rmeasure=ou2.meas.sim,
                dmeasure=ou2.meas.dens,
                skeleton.map=ou2.skel,
                t0=0
                )
coef(new.ou2) <- theta
coef(new.ou2,c("sigma.1","sigma.2","sigma.3","tau")) <- c(0,0,0,0.5)
new.ou2 <- simulate(new.ou2,seed=88737400L)


###################################################
### chunk number 38: ou2-trajmatch-calc eval=FALSE
###################################################
## tm <- traj.match(
##                  new.ou2,
##                  start=coef(new.ou2),
##                  est=c("alpha.2","alpha.3","tau","x1.0","x2.0"),
##                  method="Nelder-Mead",
##                  maxit=1000,
##                  reltol=1e-8
##                  )


###################################################
### chunk number 39: ou2-trajmatch-eval
###################################################
binary.file <- "ou2-trajmatch.rda"
if (file.exists(binary.file)) {
  load(binary.file)
} else {
tm <- traj.match(
                 new.ou2,
                 start=coef(new.ou2),
                 est=c("alpha.2","alpha.3","tau","x1.0","x2.0"),
                 method="Nelder-Mead",
                 maxit=1000,
                 reltol=1e-8
                 )
  save(new.ou2,tm,file=binary.file)
}


###################################################
### chunk number 40: trajmatch-plot
###################################################
op <- par(mfrow=c(2,1),mar=c(3,3,0,0),mgp=c(2,1,0),bty='l')
plot(time(new.ou2),data.array(new.ou2,"y1"),xlab="",xaxt='n',ylab=expression(y[1]))
x <- trajectory(new.ou2,params=tm$params)
lines(time(new.ou2),x["x1",1,-1],lwd=2)
plot(time(new.ou2),data.array(new.ou2,"y2"),xlab="time",ylab=expression(y[2]))
lines(time(new.ou2),x["x2",1,-1],lwd=2)
par(op)


###################################################
### chunk number 41: seas-basis
###################################################
tbasis <- seq(0,25,by=1/52)
basis <- periodic.bspline.basis(tbasis,nbasis=3)
colnames(basis) <- paste("seas",1:3,sep='')


###################################################
### chunk number 42: sir-proc-sim-def
###################################################
sir.proc.sim <- function (x, t, params, covars, delta.t, ...) {
  params <- exp(params)
  with(
       as.list(c(x,params,covars)),
       {
         beta <- exp(sum(log(c(beta1,beta2,beta3))*c(seas1,seas2,seas3)))
         beta.var <- beta.sd^2
         dW <- if (beta.var>0) 
           rgamma(n=1,shape=delta.t/beta.var,scale=beta.var) 
         else 
           delta.t
         foi <- (iota+beta*I*dW/delta.t)/pop
         trans <- c(
                    rpois(n=1,lambda=mu*pop*delta.t),
                    reulermultinom(n=1,size=S,rate=c(foi,mu),dt=delta.t),
                    reulermultinom(n=1,size=I,rate=c(gamma,mu),dt=delta.t),
                    reulermultinom(n=1,size=R,rate=c(mu),dt=delta.t)
                    )
         c(
           S=S+trans[1]-trans[2]-trans[3],
           I=I+trans[2]-trans[4]-trans[5],
           R=R+trans[4]-trans[6],
           cases=cases+trans[4],
           W=if (beta.sd>0) W+(dW-delta.t)/beta.sd else W
           )
       }
       )
}


###################################################
### chunk number 43: sir-pomp-def
###################################################
pomp(
     data=data.frame(
       time=seq(1/52,4,by=1/52),
       reports=NA
       ),
     times="time",
     t0=0,
     tcovar=tbasis,
     covar=basis,
     rprocess=euler.sim(
       step.fun=sir.proc.sim,
       delta.t=1/52/20
       ),
     measurement.model=reports~binom(size=cases,prob=exp(rho)),
     zeronames=c("cases"),
     initializer=function(params, t0, comp.names, ...){
       p <- exp(params)
       snames <- c("S","I","R","cases","W")
       fracs <- p[paste(comp.names,"0",sep=".")]
       x0 <- numeric(length(snames))
       names(x0) <- snames
       x0[comp.names] <- round(p['pop']*fracs/sum(fracs))
       x0
     },
     comp.names=c("S","I","R")
     ) -> sir


###################################################
### chunk number 44: sir-params
###################################################
theta <- c(
           gamma=26,mu=1/50,iota=10,
           beta1=1200,beta2=1500,beta3=900,
           beta.sd=1e-2,
           pop=5e5,
           rho=0.6,
           S.0=26/1200,I.0=0.001,R.0=1-0.001-26/1200
           )


###################################################
### chunk number 45: sir-first-sim
###################################################
sir <- simulate(sir,nsim=1,params=log(theta),seed=329348545L)


###################################################
### chunk number 46: sir-plot
###################################################
data(euler.sir)
sir <- simulate(euler.sir,nsim=1,params=c(log(theta),nbasis=3,period=1,degree=3),seed=329348545L)
plot(sir)


###################################################
### chunk number 47: restore-opts
###################################################
options(glop)


